from unittest import mock

import pytest
from django.test import TestCase
from rest_framework import serializers
from rest_framework.serializers import ListSerializer, Serializer

from ambient_toolbox.drf.fields import RecursiveField
from testapp.models import ModelWithFkToSelf, ModelWithOneToOneToSelf


class TestManyTrueSerializer(serializers.ModelSerializer):
    children = RecursiveField(many=True)

    class Meta:
        model = ModelWithFkToSelf
        fields = (
            "id",
            "children",
        )


@pytest.mark.skip
class TestManyFalseSerializer(serializers.ModelSerializer):
    peer = RecursiveField()

    class Meta:
        model = ModelWithOneToOneToSelf
        fields = (
            "id",
            "peer",
        )


class RecursiveFieldTest(TestCase):
    @mock.patch.object(Serializer, "update")
    def test_update_super_called(self, mocked_update):
        field = RecursiveField()
        field.update(instance=None, validated_data={})

        mocked_update.assert_called_once_with(None, {})

    @mock.patch.object(Serializer, "create")
    def test_create_super_called(self, mocked_create):
        field = RecursiveField()
        field.create(validated_data={})

        mocked_create.assert_called_once_with({})

    def test_many_true_regular(self):
        serializer = TestManyTrueSerializer()

        self.assertIn("children", serializer.fields)
        self.assertIsInstance(serializer.fields["children"], ListSerializer)
        self.assertIsInstance(serializer.fields["children"].child, RecursiveField)

    def test_many_true_representation(self):
        mwfts_1 = ModelWithFkToSelf.objects.create(parent=None)
        mwfts_2 = ModelWithFkToSelf.objects.create(parent=mwfts_1)

        serializer = TestManyTrueSerializer(instance=mwfts_1)
        representation = serializer.to_representation(instance=mwfts_1)

        self.assertIsInstance(representation, dict)
        self.assertIn("children", representation)
        self.assertEqual(len(representation["children"]), 1)
        self.assertEqual(representation["children"][0]["id"], mwfts_2.id)
        self.assertEqual(representation["children"][0]["children"], [])

    def test_many_false_regular(self):
        serializer = TestManyFalseSerializer()

        self.assertIn("peer", serializer.fields)
        self.assertIsInstance(serializer.fields["peer"], RecursiveField)

    def test_many_false_representation(self):
        mwotos_no_peer = ModelWithOneToOneToSelf.objects.create(peer=None)
        mwotos_has_peer = ModelWithOneToOneToSelf.objects.create(peer=mwotos_no_peer)

        serializer = TestManyFalseSerializer(instance=mwotos_has_peer)
        representation = serializer.to_representation(instance=mwotos_has_peer)

        self.assertIsInstance(representation, dict)
        self.assertIn("peer", representation)
        self.assertEqual(len(representation["peer"]), 2)
        self.assertEqual(representation["peer"]["id"], mwotos_no_peer.id)
        self.assertEqual(representation["peer"]["peer"], None)



from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout
from django import forms
from django.test import TestCase

from ambient_toolbox.admin.views.forms import AdminCrispyForm


class TestForm(AdminCrispyForm):
    my_field = forms.Field()


class AdminFormTest(TestCase):
    def test_admin_crispy_form_regular(self):
        # Form provides mostly styling, so we just validate that it renders
        form = TestForm()

        self.assertIsInstance(form.helper, FormHelper)
        self.assertIsInstance(form.helper.layout, Layout)
        self.assertEqual(form.helper.form_method, "post")



import os
import sys
from pathlib import Path
from typing import Union

from django.conf import settings

from ambient_toolbox.tests.structure_validator import settings as toolbox_settings


class TestStructureValidator:
    file_whitelist: list
    issue_list: list

    def __init__(self):
        self.file_whitelist = self._get_file_whitelist()
        self.issue_list = []

    @staticmethod
    def _get_file_whitelist() -> list:
        default_whitelist = ["__init__"]
        try:
            return default_whitelist + settings.TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST
        except AttributeError:
            return default_whitelist + toolbox_settings.TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST

    @staticmethod
    def _get_base_dir() -> Union[Path, str]:
        try:
            return settings.TEST_STRUCTURE_VALIDATOR_BASE_DIR
        except AttributeError:
            return toolbox_settings.TEST_STRUCTURE_VALIDATOR_BASE_DIR

    @staticmethod
    def _get_base_app_name() -> str:
        try:
            return settings.TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME
        except AttributeError:
            return toolbox_settings.TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME

    @staticmethod
    def _get_ignored_directory_list() -> list:
        default_dir_list = ["__pycache__"]
        try:
            return default_dir_list + settings.TEST_STRUCTURE_VALIDATOR_IGNORED_DIRECTORY_LIST
        except AttributeError:
            return default_dir_list + toolbox_settings.TEST_STRUCTURE_VALIDATOR_IGNORED_DIRECTORY_LIST

    @staticmethod
    def _get_app_list() -> Union[list, tuple]:
        try:
            return settings.TEST_STRUCTURE_VALIDATOR_APP_LIST
        except AttributeError:
            return toolbox_settings.TEST_STRUCTURE_VALIDATOR_APP_LIST

    def _check_missing_test_prefix(self, *, root: str, file: str, filename: str, extension: str) -> bool:
        if extension == ".py" and not filename[0:5] == "test_" and filename not in self.file_whitelist:
            file_path = f"{root}\\{file}".replace("\\", "/")
            self.issue_list.append(f'Python file without "test_" prefix found: {file_path!r}.')
            return False
        return True

    def _check_missing_init(self, *, root: str, init_found: bool, number_of_py_files: int) -> bool:
        if not init_found and number_of_py_files > 0:
            path = root.replace("\\", "/")
            self.issue_list.append(f"__init__.py missing in {path!r}.")
            return False
        return True

    def _build_path_to_test_package(self, app: str) -> Path:
        return self._get_base_dir() / Path(app.replace(".", "/")) / "tests"

    def process(self) -> None:
        backend_package = self._get_base_app_name()
        app_list = self._get_app_list()

        for app in app_list:
            if not app.startswith(backend_package):
                continue
            app_path = self._build_path_to_test_package(app=app)
            for root, dirs, files in os.walk(app_path):
                cleaned_root = root.replace("\\", "/")
                print(f"Inspecting {cleaned_root!r}...")
                init_found = False
                number_of_py_files = 0

                for excluded_dir in self._get_ignored_directory_list():
                    try:
                        dirs.remove(excluded_dir)
                    except ValueError:
                        pass

                for file in files:
                    filename = file[:-3]
                    extension = file[-3:]

                    if filename == "__init__":
                        init_found = True

                    if extension == ".py":
                        number_of_py_files += 1

                    # Check for missing test prefix
                    self._check_missing_test_prefix(root=root, file=file, filename=filename, extension=extension)

                # Check for missing init file
                self._check_missing_init(root=root, init_found=init_found, number_of_py_files=number_of_py_files)

        number_of_issues = len(self.issue_list)

        if number_of_issues:
            print("=======================")
            print("Errors found:")

            for issue in self.issue_list:
                print(f"- {issue}")

        print("=======================")

        if number_of_issues:
            print(f"Checking test structure failed with {number_of_issues} issue(s).")
            sys.exit(1)
        else:
            print("0 issues detected. Yeah!")



from unittest import mock

from django.test import TestCase

from testapp.models import ModelWithCleanMixin


class CleanOnSaveMixinTest(TestCase):
    def test_clean_regular(self):
        obj = ModelWithCleanMixin()
        self.assertIsNone(obj.save())

    def test_clean_is_called(self):
        obj = ModelWithCleanMixin()
        with mock.patch.object(obj, "clean") as mocked_method:
            obj.save()

        mocked_method.assert_called_once()



from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

from ambient_toolbox.managers import GloballyVisibleQuerySet
from ambient_toolbox.mixins.bleacher import BleacherMixin
from ambient_toolbox.mixins.models import PermissionModelMixin, SaveWithoutSignalsMixin
from ambient_toolbox.mixins.validation import CleanOnSaveMixin
from ambient_toolbox.models import CommonInfo
from testapp.managers import ModelWithSelectorQuerySet
from testapp.selectors import ModelWithSelectorGloballyVisibleSelector


class MySingleSignalModel(models.Model):
    value = models.PositiveIntegerField(default=0)

    objects = GloballyVisibleQuerySet.as_manager()

    def __str__(self):
        return str(self.value)


class ForeignKeyRelatedModel(models.Model):
    single_signal = models.ForeignKey(
        MySingleSignalModel, on_delete=models.CASCADE, related_name="foreign_key_related_models"
    )

    objects = GloballyVisibleQuerySet.as_manager()

    def __str__(self):
        return str(self.id)


@receiver(pre_save, sender=MySingleSignalModel)
def increase_value_no_dispatch_uid(sender, instance, **kwargs):
    instance.value += 1


class MyMultipleSignalModel(models.Model):
    value = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


@receiver(pre_save, sender=MyMultipleSignalModel, dispatch_uid="test.mysinglesignalmodel.increase_value_with_uuid")
def increase_value_with_dispatch_uid(sender, instance, **kwargs):
    instance.value += 1


@receiver(post_save, sender=MyMultipleSignalModel)
def send_email(sender, instance, **kwargs):
    msg = EmailMultiAlternatives(
        "Test Mail", "I am content", from_email="test@example.com", to=["random.dude@example.com"]
    )
    msg.send()


class CommonInfoBasedModel(CommonInfo):
    value = models.PositiveIntegerField(default=0)
    value_b = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


class ModelWithSelector(models.Model):
    value = models.PositiveIntegerField(default=0)

    objects = ModelWithSelectorQuerySet.as_manager()
    selectors = ModelWithSelectorGloballyVisibleSelector()

    def __str__(self):
        return str(self.value)


class ModelWithFkToSelf(models.Model):
    parent = models.ForeignKey("self", blank=True, null=True, related_name="children", on_delete=models.CASCADE)

    def __str__(self):
        return str(self.id)


class ModelWithOneToOneToSelf(models.Model):
    peer = models.OneToOneField("self", blank=True, null=True, related_name="related_peer", on_delete=models.CASCADE)

    def __str__(self):
        return str(self.id)


class ModelWithCleanMixin(CleanOnSaveMixin, models.Model):
    def __str__(self):
        return str(self.id)

    def clean(self):
        return True


class MyPermissionModelMixin(PermissionModelMixin, models.Model):
    def __str__(self):
        return str(self.id)


class BleacherMixinModel(BleacherMixin, models.Model):
    BLEACH_FIELD_LIST = ["content"]

    content = models.CharField(max_length=50)

    def __str__(self):
        return self.content


class ModelWithSaveWithoutSignalsMixin(SaveWithoutSignalsMixin, models.Model):
    value = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


@receiver(pre_save, sender=ModelWithSaveWithoutSignalsMixin)
def increase_value_on_pre_save(sender, instance, **kwargs):
    instance.value += 1



from sentry_sdk.integrations.logging import ignore_logger


def ignore_graphene_logger():
    """
    A utils function that can be called to ignore a graphene logger that logs errors as strings instead of errors.
    This leads to a loss of Sentry's features. Instead, you can report the original exception.

    Test for:
        * sentry_sdk >= 0.13.0
    """
    ignore_logger("graphql.execution.utils")



import re

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


class SpecialCharValidator:
    """
    The password must contain at least one special character (@#$%!^&*)
    """

    def validate(self, password, user=None):
        if not re.findall("[@#$%!^&*]", password):
            raise ValidationError(
                _('The password has to contain one of the following special characters: "@#$%!^&*"'),
                code="password_no_symbol",
            )

    def get_help_text(self):
        return _('The password has to contain one of the following special characters: "@#$%!^&*"')



from django.test import TestCase

from testapp.models import ModelWithSaveWithoutSignalsMixin


class SaveWithoutSignalsTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.instance = ModelWithSaveWithoutSignalsMixin.objects.create()

    def test_signals_are_executed_on_normal_save_call(self):
        value_before = self.instance.value

        self.instance.save()
        self.instance.refresh_from_db()

        self.assertEqual(value_before + 1, self.instance.value)

    def test_signals_are_not_executed_on_save_without_signals_call(self):
        value_before = self.instance.value

        self.instance.save_without_signals()
        self.instance.refresh_from_db()

        self.assertEqual(value_before, self.instance.value)

    def test_signals_are_not_executed_on_save_without_signals_call_but_then_reexecuted_on_normal_save_call(self):
        value_before = self.instance.value

        self.instance.save_without_signals()
        self.instance.refresh_from_db()

        self.assertEqual(value_before, self.instance.value)

        self.instance.save()

        self.assertEqual(value_before + 1, self.instance.value)



from django.test import TestCase

from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration, PermissionModelDeclaration


class PermissionFixtureDeclarationTest(TestCase):
    def test_permission_model_declaration_regular(self):
        permission = PermissionModelDeclaration(
            app_label="my_app",
            codename_list=["view_mymodel"],
            model="mymodel",
        )

        self.assertEqual(permission.app_label, "my_app")
        self.assertEqual(permission.codename_list, ["view_mymodel"])
        self.assertEqual(permission.model, "mymodel")

    def test_group_permission_declaration_regular(self):
        permission = PermissionModelDeclaration(
            app_label="my_app",
            codename_list=["view_mymodel"],
            model="mymodel",
        )
        group = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[permission],
        )

        self.assertEqual(group.name, "my_group")
        self.assertEqual(group.permission_list, [permission])



from os.path import isfile
from pathlib import Path
from typing import Optional

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """
    Wrapper command for "makemessages" to remove "POT-Creation-Date:" which will cause countless merge conflicts.
    """

    def add_arguments(self, parser):
        parser.add_argument("--lang", type=str)

    def detect_single_translation_language(self) -> Optional[str]:
        """
        Returns the language code of the projects' translation if there is exactly one.
        Returns None otherwise.
        """
        locale_path = settings.LOCALE_PATHS[0]

        # We don't bother with odd configurations for a convenience feature
        if not isinstance(locale_path, Path):
            return None

        dir_list: list[Path] = [directory for directory in locale_path.iterdir() if directory.is_dir()]

        if len(dir_list) == 1:
            return dir_list[0].name
        return None

    def handle(self, *args, **options):
        try:
            language = options["lang"].strip('"')
        except AttributeError as e:
            language = self.detect_single_translation_language()

            if not language:
                raise RuntimeError('Please provide a language with the "--lang" parameter.') from e

        call_command("makemessages", "-l", language, "--no-location", "--no-wrap")

        file_path = f"{settings.LOCALE_PATHS[0]}/{language}/LC_MESSAGES/django.po"
        if not isfile(file_path):
            raise RuntimeError(f'PO file not found at "{file_path}".')

        with open(file_path, encoding="utf-8") as po_file:
            content_list = po_file.readlines()

        modified_content = [line for line in content_list if not line.startswith('"POT-Creation-Date:')]

        with open(file_path, "w", encoding="utf-8") as po_file:
            po_file.writelines(modified_content)



from typing import Optional, Union

from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages import get_messages
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.test import RequestFactory
from django.utils.translation import gettext_lazy as _


class ClassBasedViewTestMixin:
    """
    Helper mixin for easy testing of Django views
    """

    factory_class = RequestFactory
    view_class = None

    @staticmethod
    def _authentication(request, user):
        request.user = user if user else AnonymousUser()

    def _get_response(self, method, user, data, url_params=None, *args, **kwargs):
        """Returns response."""
        # Catch case that URL does not get any params passed to
        if not url_params:
            url_params = {}

        # Create request
        factory = self.factory_class()
        req_kwargs = {}
        if data:
            req_kwargs.update({"data": data})
        request = getattr(factory, method)("/", **req_kwargs)

        # Annotate a request object with a session
        middleware = SessionMiddleware(get_response=HttpResponse(status=200))
        middleware.process_request(request)
        request.session.save()

        # Authenticate user
        self._authentication(request, user)

        # Setup messages
        messages = FallbackStorage(request)
        request._messages = messages

        # Call view
        return self.view_class.as_view()(request, *args, **url_params)

    def get(self, user=None, data=None, url_params=None, *args, **kwargs):
        """Returns response for a GET request."""
        return self._get_response("get", user, data, url_params, *args, **kwargs)

    def post(self, user=None, data=None, url_params=None, *args, **kwargs):
        """Returns response for a POST request."""
        return self._get_response("post", user, data, url_params, *args, **kwargs)

    def delete(self, user=None, data=None, url_params=None, *args, **kwargs):
        """Returns response for a DELETE request."""
        return self._get_response("delete", user, data, url_params, *args, **kwargs)


class RequestProviderMixin:
    """
    Mixin that provides a method which returns a django request.
    Supposed to be used in unittests that require a request object.
    Also adds the message middleware so django messages can be tested.
    """

    @staticmethod
    def get_request(
        user: Union[AbstractBaseUser, AnonymousUser, None] = None, method: str = "GET", url: Optional[str] = None
    ):
        """
        Creates and returns a django request.
        """
        # Determine URL
        url = url if url else "/"

        # Create test request
        factory = RequestFactory()
        request = factory.get(url)

        # Set user object if it is of a valid type
        if user is None or isinstance(user, AbstractBaseUser) or isinstance(user, AnonymousUser):
            request.user = user
        else:
            raise ValueError(_("Please pass a user object to RequestProviderMixin."))

        # Annotate a request object with a session
        session_middleware = SessionMiddleware(get_response=HttpResponse(status=200))
        session_middleware.process_request(request)
        request.session.save()

        # Annotate a request object with messages
        message_middleware = MessageMiddleware(get_response=HttpResponse(status=200))
        message_middleware.process_request(request)
        request.session.save()

        # Set request method
        request.method = method

        return request


class DjangoMessagingFrameworkTestMixin:
    """
    Mixin to enable test cases to easily check for messages in the request.
    This test is made for the Django Messaging Framework.
    """

    def assert_full_message_in_request(self, request: WSGIRequest, message: str) -> None:
        messages = list(get_messages(request))

        self.assertGreater(len(messages), 0, "The request doesn't contain any messages.")
        message_found = False
        for request_message in messages:
            if str(request_message) == message:
                message_found = True

        self.assertTrue(message_found, f"Message {message!r} not found in request.")

    def assert_partial_message_in_request(self, request: WSGIRequest, message: str) -> None:
        messages = list(get_messages(request))

        self.assertGreater(len(messages), 0, "The request doesn't contain any messages.")

        message_found = False
        for request_message in messages:
            if message in str(request_message):
                message_found = True

        self.assertTrue(message_found, f"Message part {message!r} not found in request.")

    def assert_message_not_in_request(self, request: WSGIRequest, message: str) -> None:
        messages = list(get_messages(request))

        message_found = False
        for request_message in messages:
            if str(request_message) == message:
                message_found = True

        self.assertFalse(message_found, f"Message {message!r} not found in request.")



from collections import OrderedDict

from django.test import TestCase

from ambient_toolbox.utils import get_key_from_tuple_by_value, get_namedtuple_choices, get_value_from_tuple_by_key


class UtilsNamedTupleTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.MY_CHOICE_ONE = 1
        cls.MY_CHOICE_TWO = 2
        cls.MY_CHOICE_THREE = 3

        cls.MY_CHOICE_LIST = (
            (cls.MY_CHOICE_ONE, "Choice 1"),
            (cls.MY_CHOICE_TWO, "Choice 2"),
            (cls.MY_CHOICE_THREE, "Choice 3"),
        )

        cls.colors_choices = get_namedtuple_choices(
            "COLORS",
            (
                (1, "black", "Black"),
                (2, "white", "White"),
            ),
        )

    def test_get_namedtuple_choices_regular(self):
        self.assertEqual(self.colors_choices.black, 1)
        self.assertEqual(self.colors_choices.white, 2)

    def test_get_namedtuple_choices_get_choices_regular(self):
        self.assertEqual(self.colors_choices.get_choices(), [(1, "Black"), (2, "White")])

    def test_get_namedtuple_choices_get_choices_dict_regular(self):
        self.assertEqual(self.colors_choices.get_choices_dict(), OrderedDict([(1, "Black"), (2, "White")]))

    def test_get_namedtuple_choices_get_all_regular(self):
        for index, color in enumerate(self.colors_choices.get_all()):
            expected_tuple = "invalid_data"
            if index == 0:
                expected_tuple = (1, "black", "Black")
            elif index == 1:
                expected_tuple = (2, "white", "White")
            self.assertEqual(color, expected_tuple)

    def test_get_namedtuple_choices_get_choices_tuple_regular(self):
        self.assertEqual(self.colors_choices.get_choices_tuple(), ((1, "black", "Black"), (2, "white", "White")))

    def test_get_namedtuple_choices_get_values_regular(self):
        self.assertEqual(self.colors_choices.get_values(), [1, 2])

    def test_get_namedtuple_choices_get_value_by_name_regular(self):
        self.assertEqual(self.colors_choices.get_value_by_name("black"), 1)
        self.assertEqual(self.colors_choices.get_value_by_name("white"), 2)
        self.assertFalse(self.colors_choices.get_value_by_name("no-existing"))

    def test_get_namedtuple_choices_get_desc_by_value_regular(self):
        self.assertEqual(self.colors_choices.get_desc_by_value(1), "Black")
        self.assertEqual(self.colors_choices.get_desc_by_value(2), "White")
        self.assertFalse(self.colors_choices.get_desc_by_value(-1))

    def test_get_namedtuple_choices_get_name_by_value_regular(self):
        self.assertEqual(self.colors_choices.get_name_by_value(1), "black")
        self.assertEqual(self.colors_choices.get_name_by_value(2), "white")
        self.assertFalse(self.colors_choices.get_name_by_value(-1))

    def test_get_namedtuple_choices_is_valid_regular(self):
        self.assertTrue(self.colors_choices.is_valid(1))
        self.assertTrue(self.colors_choices.is_valid(2))
        self.assertFalse(self.colors_choices.is_valid(-1))

    def test_get_value_from_tuple_by_key_found(self):
        self.assertEqual(get_value_from_tuple_by_key(self.MY_CHOICE_LIST, self.MY_CHOICE_TWO), "Choice 2")

    def test_get_value_from_tuple_by_key_not_found(self):
        self.assertEqual(get_value_from_tuple_by_key(self.MY_CHOICE_LIST, 99), "-")

    def test_get_key_from_tuple_by_value_found(self):
        self.assertEqual(get_key_from_tuple_by_value(self.MY_CHOICE_LIST, "Choice 2"), self.MY_CHOICE_TWO)

    def test_get_key_from_tuple_by_value_not_found(self):
        self.assertEqual(get_key_from_tuple_by_value(self.MY_CHOICE_LIST, "Something odd"), "-")

    def test_get_values_case_is_instance(self):
        choices = get_namedtuple_choices(
            "TestChoices",
            (
                (0, "zero", "Zero"),
                (1, "one", "One"),
                ([2, 3], "two_three", "Two Three"),
            ),
        )

        # Überprüfen, ob die Methode get_values korrekt funktioniert
        self.assertEqual(choices.get_values(), [0, 1, 2, 3])



from unittest.mock import PropertyMock, patch

from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time

from testapp.models import CommonInfoBasedModel


class CommonInfoTest(TestCase):
    @freeze_time("2022-06-26 10:00")
    def test_save_created_at_set(self):
        obj = CommonInfoBasedModel.objects.create(value=1, value_b=1)
        obj.created_at = None
        obj.save()

        self.assertEqual(obj.created_at, timezone.now())

    @freeze_time("2022-06-26 10:00")
    def test_save_update_fields_common_fields_set(self):
        with freeze_time("2020-09-19"):
            obj = CommonInfoBasedModel.objects.create(value=1, value_b=1)
        obj.value = 2
        obj.value_b = 999

        # Django's Model.save() can be called with positional args, so we should support this as well.
        args = (
            False,  # default for force_insert
            False,  # default for force_update
            None,  # default for using
            (x for x in ["value"]),  # update_fields is supposed to accept any Iterable[str]
        )
        obj.save(*args)

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.value_b, 1, "value_b should not have changed")
        self.assertEqual(obj.lastmodified_at.year, 2022)
        self.assertEqual(obj.lastmodified_at.month, 6)
        self.assertEqual(obj.lastmodified_at.day, 26)
        self.assertEqual(obj.lastmodified_at.hour, 10)

    @patch("testapp.models.CommonInfoBasedModel.ALWAYS_UPDATE_FIELDS", new_callable=PropertyMock)
    @freeze_time("2022-06-26 10:00")
    def test_save_update_fields_common_fields_set_without_always_update(self, always_update_mock):
        always_update_mock.return_value = False
        with freeze_time("2020-09-19"):
            obj = CommonInfoBasedModel.objects.create(value=1)
        obj.value = 2
        obj.save(update_fields=("value",))

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.lastmodified_at.year, 2020)
        self.assertEqual(obj.lastmodified_at.month, 9)
        self.assertEqual(obj.lastmodified_at.day, 19)
        self.assertEqual(obj.lastmodified_at.hour, 0)

    @freeze_time("2022-06-26 10:00")
    def test_save_common_fields_set_without_update_fields(self):
        with freeze_time("2020-09-19"):
            obj = CommonInfoBasedModel.objects.create(value=1)
        obj.value = 2
        obj.save()

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.lastmodified_at.year, 2022)
        self.assertEqual(obj.lastmodified_at.month, 6)
        self.assertEqual(obj.lastmodified_at.day, 26)
        self.assertEqual(obj.lastmodified_at.hour, 10)
